///////////////////////////////////////////////////////////////////////////////
// Circuit Playrgound How Cold Is It - Logger Version
//
// Author: Carter Nelson
// MIT License (https://opensource.org/licenses/MIT)

#include <Adafruit_CircuitPlayground.h>
#include <EEPROM.h>

// OK to change these
#define LOG_WRAP    false // Set log wrapping
#define LOG_RATE    1000  // Set log rate in milliseconds

// Don't change these
#define LOG_INFO    0     // EEPROM address for log meta data
#define LOG_START   2     //    "      "     "   "  start
#define LOG_END     1022  //    "      "     "   "  end
#define DISP_RATE   30    // NeoPixel display update rate
#define BEAT_RATE   1000  // Red LED heartbeat rate

bool logIsFull;           // True if log is full
uint8_t logFormat;        // Format of log output 
uint16_t logIndex;        // Index of next log location 
bool logEnabled;          // Enable/disable logging

uint32_t startDelay;      // Start time for log delay
uint32_t delayTime;       // Amount of delay
uint32_t lastUpdate;      // Last display update time
uint32_t lastBeat;        // Last heart beat update time
bool beat;                // And the beat goes on...

//  Log Meta Info Layout
//  |-------byte1----------||-------byte2----------|
//  [7][6][5][4][3][2][1][0][7][6][5][4][3][2][1][0]
//  |F||FORMAT-||----||--------LOGINDEX------------|
///////////////////////////////////////////////////////////////////////////////
void logReadInfo() {
  uint8_t byte1 = EEPROM.read(LOG_INFO);
  uint8_t byte2 = EEPROM.read(LOG_INFO+1);

  logIsFull = bool((byte1 >> 7) & 0x01);
  logFormat = uint8_t((byte1 >> 4) & 0x07);
  logIndex = ((byte1 & 0x03) << 8) | byte2;
}

///////////////////////////////////////////////////////////////////////////////
void logWriteInfo() {
  uint8_t byte1;
  uint8_t byte2;

  byte1 = ( ((uint8_t(logIsFull) & 0x01) << 7)  |
            ((logFormat & 0x07) << 4)           | 
            ((logIndex >> 8) & 0x03)            );

  byte2 = logIndex & 0xFF;

  EEPROM.write(LOG_INFO,   byte1);
  EEPROM.write(LOG_INFO+1, byte2);
}

///////////////////////////////////////////////////////////////////////////////
uint16_t logReadValue(uint16_t index) {
  if ((index < LOG_START) || (index > LOG_END)) {
    return NULL;
  }

  uint8_t byte1 = EEPROM.read(index++);
  uint8_t byte2 = EEPROM.read(index);

  return uint16_t((byte1 & 0xFF) << 8) | (byte2 & 0xFF); 
}

///////////////////////////////////////////////////////////////////////////////
void logWriteValue(uint16_t value) {
  if (logIsFull && !LOG_WRAP) {
    logEnabled = false;
    return;
  }
  
  if (logIndex <= LOG_END) {
    EEPROM.write(logIndex++, (value >> 8) & 0xFF);
    EEPROM.write(logIndex++, value & 0xFF);
  }
  
  if (logIndex > LOG_END) {
    logIsFull = true;
    if (LOG_WRAP) {
      logIndex = LOG_START;
    } else {
      logIndex = LOG_END+1;
    }
  }

  logWriteInfo();
}

///////////////////////////////////////////////////////////////////////////////
float logReadTemperature(uint16_t index) {
  return float(logReadValue(index))/100.0;
}

///////////////////////////////////////////////////////////////////////////////
void logWriteTemperature(float kelvin) {
  logWriteValue(uint16_t(kelvin * 100));  
}

///////////////////////////////////////////////////////////////////////////////
float tempConv(float kelvin, uint8_t tempUnit) {
  switch (tempUnit) {
    case 0:   // Celsius
      return kelvin - 273.15;
    case 1:   // FAhrenheit
      return 1.8*kelvin - 459.67;
    case 2:   // Kelvin
      return kelvin;
    case 3:   // Rankine
      return 1.8*kelvin;
  }  
}

///////////////////////////////////////////////////////////////////////////////
void logFormatChange() {
  logFormat++;
  if (logFormat > 4) logFormat = 0;  
}

///////////////////////////////////////////////////////////////////////////////
void logReset() {
  logIndex = LOG_START;
  logIsFull = false;
  logEnabled = false;
  logWriteInfo();
}

///////////////////////////////////////////////////////////////////////////////
void toggleLogging() {
  logEnabled = !logEnabled;
}

///////////////////////////////////////////////////////////////////////////////
void logDump() {
  logReadInfo();

  switch (logFormat) {
    case 0:
      logDumpTemperature(0);  break;
    case 1:
      logDumpTemperature(1);  break;
    case 2:
      logDumpSummary();       break;
    case 3:
      logDumpTable(false);    break;
    case 4:
      logDumpTable(true);     break;
    default:
      break;
  }
}

///////////////////////////////////////////////////////////////////////////////
void logDumpTemperature(uint8_t tempUnit) {
  if ((LOG_START == logIndex) && (!logIsFull)) {
    Serial.println("LOG IS EMPTY!");
    return;
  }
  
  if (logIsFull && LOG_WRAP) {
    for (int i=logIndex; i<=LOG_END; i+=sizeof(uint16_t)) {
      Serial.println(tempConv(logReadTemperature(i), tempUnit));     
    }
    for (int i=LOG_START; i<logIndex; i+=sizeof(uint16_t)) {
      Serial.println(tempConv(logReadTemperature(i), tempUnit));           
    }
  } else {
    for (int i=LOG_START; i<logIndex; i+=sizeof(uint16_t)) {
      Serial.println(tempConv(logReadTemperature(i), tempUnit));                             
    }
  }  
}

///////////////////////////////////////////////////////////////////////////////
void logDumpSummary() {
  float minT, maxT;

  Serial.println("=============");
  Serial.println(" LOG SUMMARY");
  Serial.println("=============");
  Serial.print("Full: ");
  if (logIsFull) Serial.print("yes"); else Serial.print("no");
  Serial.print("\tAllow Wrap: ");
  if (LOG_WRAP) Serial.println("yes"); else Serial.println("no");
  Serial.print("Log Start: "); Serial.print(LOG_START);
  Serial.print("\tLog End: "); Serial.print(LOG_END);
  Serial.print("\tLog Index: "); Serial.println(logIndex);
  Serial.print("Log Rate: "); Serial.println(LOG_RATE);
  if ((LOG_START == logIndex) && (!logIsFull)) {
    Serial.println("LOG IS EMPTY!");
  } else {
    Serial.print("Number of Readings: ");
    if (logIsFull) {
      Serial.println(1+(LOG_END - LOG_START)/sizeof(uint16_t));      
    } else {
      Serial.println((logIndex - LOG_START)/sizeof(uint16_t));
    }
    Serial.println("-----------");
    logGetMinMax(minT, maxT);
    Serial.print("Minimum Temp: ");
    Serial.print(tempConv(minT, 0)); Serial.print(" C  ");
    Serial.print(tempConv(minT, 1)); Serial.println(" F");
    Serial.print("Maximum Temp: ");
    Serial.print(tempConv(maxT, 0)); Serial.print(" C  ");
    Serial.print(tempConv(maxT, 1)); Serial.println(" F");
  }
  Serial.println("-----------");
  Serial.print("Current Temp: ");
  minT = CircuitPlayground.temperature() + 273.15;
  Serial.print(tempConv(minT, 0)); Serial.print(" C  ");
  Serial.print(tempConv(minT, 1)); Serial.println(" F");    
}

///////////////////////////////////////////////////////////////////////////////
void logDumpTable(bool showHeaders) {
  if ((LOG_START == logIndex) && (!logIsFull)) {
    Serial.println("LOG IS EMPTY!");
    return;
  }
  
  float kelvin;
  float secs = 0.0;
  float ds = float(LOG_RATE)/1000.0;

  if (showHeaders) {
    Serial.println("SECS DEGC DEGF DEGK DEGR");
  }
  if (logIsFull && LOG_WRAP) {
    for (int i=logIndex; i<=LOG_END; i+=sizeof(uint16_t)) {
      printTableRow(secs, logReadTemperature(i));
      secs += ds;
    }
    for (int i=LOG_START; i<logIndex; i+=sizeof(uint16_t)) {
      printTableRow(secs, logReadTemperature(i));
      secs += ds;   
    }
  } else {
    for (int i=LOG_START; i<logIndex; i+=sizeof(uint16_t)) {
      printTableRow(secs, logReadTemperature(i));
      secs += ds;   
    }
  }   
}

///////////////////////////////////////////////////////////////////////////////
void printTableRow(float secs, float kelvin) {
  Serial.print(secs); Serial.print(" ");
  Serial.print(tempConv(kelvin, 0)); Serial.print(" ");
  Serial.print(tempConv(kelvin, 1)); Serial.print(" ");
  Serial.print(tempConv(kelvin, 2)); Serial.print(" ");
  Serial.print(tempConv(kelvin, 3)); Serial.println();  
}

///////////////////////////////////////////////////////////////////////////////
void logGetMinMax(float &minTemp, float &maxTemp) {
  uint16_t index;
  float temp;
  float minT = logReadTemperature(LOG_START);
  float maxT = minT;
  
  if (logIsFull) index = LOG_END+1; else index = logIndex;
   
  for (int i=LOG_START+2; i<index; i+=sizeof(uint16_t)) {
    temp = logReadTemperature(i);
    minT = temp < minT ? temp : minT;
    maxT = temp > maxT ? temp : maxT;
  }

  minTemp = minT;
  maxTemp = maxT;
}

///////////////////////////////////////////////////////////////////////////////
void updateDisplay() {
  if (millis() - lastUpdate < DISP_RATE) return;
  lastUpdate = millis();
  
  CircuitPlayground.clearPixels();
  
  if (CircuitPlayground.slideSwitch()) {
    //----------------------| LOGGING MODE |-----------------------------------
    for (int p=0; p<5; p++) CircuitPlayground.setPixelColor(p, 0x00FF00);
    if (!logEnabled) {
      CircuitPlayground.redLED(false);
      CircuitPlayground.setPixelColor(2, 0xFF0000);
    }
    int pindex = 5 + 5 * (logIndex - LOG_START)/(LOG_END - LOG_START);
    if (logIsFull) {
      for (int p=5; p<10; p++) CircuitPlayground.setPixelColor(p, 0xFF0000);
      if (LOG_WRAP) CircuitPlayground.setPixelColor(pindex, 0xFFFFFF);
    } else if (logIndex != LOG_START) {
      for (int p=5; p<=pindex; p++) CircuitPlayground.setPixelColor(p, 0xFFFFFF);  
    }
  
  } else {
    //----------------------| READOUT MODE |-----------------------------------
    CircuitPlayground.redLED(false);    
    for (int p=0; p<5; p++) CircuitPlayground.setPixelColor(p, 0x00FFFF);
    CircuitPlayground.setPixelColor(9-logFormat, 0xFF00FF);    
  }

}

///////////////////////////////////////////////////////////////////////////////
void heartBeat() {
  if (millis() - lastBeat < BEAT_RATE) return;
  lastBeat = millis();
  beat = !beat;
  CircuitPlayground.redLED(beat);
}

///////////////////////////////////////////////////////////////////////////////
void setup() {
  Serial.begin(9600);

  CircuitPlayground.begin();

  logEnabled = false;
  lastUpdate = millis();
  lastBeat = millis();
  beat = false;
  logReadInfo();
}

///////////////////////////////////////////////////////////////////////////////
void loop() {
  updateDisplay();
  if (CircuitPlayground.slideSwitch()) {
    
    //----------------------| LOGGING MODE |-----------------------------------
    if (logEnabled) {
      logWriteTemperature(CircuitPlayground.temperature() + 273.15);
      delayTime = LOG_RATE;
    } else {
      delayTime = 250;
    }
    startDelay = millis();
    while (millis() - startDelay < delayTime) {
      if (!CircuitPlayground.slideSwitch()) break;
      if (CircuitPlayground.leftButton() && CircuitPlayground.rightButton()) {
        for (int p=0; p<10; p++) CircuitPlayground.setPixelColor(p, 0xFF0000);
        logReset();
        delay(250);
        break;
      }
      if (CircuitPlayground.leftButton()) {
        toggleLogging();
        delay(250);
        break;
      }
      if (logEnabled) heartBeat();
      updateDisplay();            
    }
    
  } else {
    
    //----------------------| READOUT MODE |-----------------------------------
    logEnabled = false;
    if (CircuitPlayground.leftButton()) {
      logDump();
      delay(250);
    }
    if (CircuitPlayground.rightButton()) {
      logFormatChange();
      updateDisplay();
      logWriteInfo();
      delay(250);
    }

  }
}
